# Copyright (c) 2020 mingruimingrui
# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the MIT-style license found in the
# LICENSE file in the root directory of this source tree.

from laser.bin import encode_texts, filter_parallel

__all__ = ['encode_texts', 'filter_parallel']
